(*****************************************************************************)
(*                                                                           *)
(* MIT License                                                               *)
(* Copyright (c) 2022 Nomadic Labs <contact@nomadic-labs.com>                *)
(*                                                                           *)
(* Permission is hereby granted, free of charge, to any person obtaining a   *)
(* copy of this software and associated documentation files (the "Software"),*)
(* to deal in the Software without restriction, including without limitation *)
(* the rights to use, copy, modify, merge, publish, distribute, sublicense,  *)
(* and/or sell copies of the Software, and to permit persons to whom the     *)
(* Software is furnished to do so, subject to the following conditions:      *)
(*                                                                           *)
(* The above copyright notice and this permission notice shall be included   *)
(* in all copies or substantial portions of the Software.                    *)
(*                                                                           *)
(* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR*)
(* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *)
(* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL   *)
(* THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER*)
(* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING   *)
(* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER       *)
(* DEALINGS IN THE SOFTWARE.                                                 *)
(*                                                                           *)
(*****************************************************************************)

module Stubs = struct
  type t

  external allocate : int -> t = "caml_polynomial_srs_allocate_stubs"
  (** [allocate n] allocates a custom block to save a SRS of size [n] *)

  external load_from_file : t -> string -> int -> int -> bool
    = "caml_polynomial_srs_load_from_file_stubs"

  external of_array : t -> Bls12_381.G1.t array -> int -> unit
    = "caml_polynomial_srs_of_array_stubs"
    [@@noalloc]
  (** [of_array srs jacobian_pts length] feeds the SRS
      [srs] with the SRS given in jacobian coordinates *)

  external to_array : Bls12_381.G1.t array -> t -> int -> unit
    = "caml_polynomial_srs_to_array_stubs"
    [@@noalloc]
  (** [to_array jacobian_pts_array srs length]
      extracts the SRS points from [srs] and feed it in
      [jacobian_pts_array]
  *)

  external pippenger :
    Bls12_381.G1.t -> t -> Carray.Stubs.fr_array -> int -> int -> int -> unit
    = "caml_polynomial_pippenger_stubs_bytecode" "caml_polynomial_pippenger_stubs"
  (** [pippenger res srs fr_array start len] uses pippenger to compute multi
      scalar exponentiation *)
end

module type S = sig
  type polynomial

  type t
  (** A context to be used with {!pippenger} *)

  val of_array : Bls12_381.G1.t array -> t
  val to_array : t -> Bls12_381.G1.t array

  val size : t -> int
  (** Returns the pippenger ctxt size, i.e. the number of elements the context
      is supposed to be called with *)

  val load_from_file : string -> ?offset:int -> int -> t
  (** [srs_load_from_file path ~offset size] reads [size] points of g1 in affine
      compressed format from the file at [path] starting at [offset]. *)

  val pippenger : ?start:int -> ?len:int -> t -> polynomial -> Bls12_381.G1.t
  (** [pippenger ctxt poly] computes the multiscalar exponentiation using the
      SRS saved in the context and the coefficients of the given polynomial *)
end

module M = struct
  type polynomial = Polynomial_c.Polynomial_unsafe.t
  type t = Stubs.t * int

  let of_array a =
    let length = Array.length a in
    let srs = Stubs.allocate length in
    Stubs.of_array srs a length;
    (srs, length)

  let to_array (srs, n) =
    let pts = Array.init n (fun _ -> Bls12_381.G1.(copy one)) in
    Stubs.to_array pts srs n;
    pts

  let load_from_file path ?(offset = 0) n =
    let srs = Stubs.allocate n in
    let success = Stubs.load_from_file srs path n offset in
    if success then (srs, n)
    else
      raise
      @@ Invalid_argument
           (Format.sprintf
              "The path %s is either invalid or it does not contain %d \
               elements from offset %d"
              path n offset)

  let size (_, n) = n

  let pippenger ?(start = 0) ?len (srs, ctxt_size) poly =
    let coefficients, n = Polynomial_c.Polynomial_unsafe.to_carray poly in
    let l = min ctxt_size n in
    let len = Option.value ~default:(l - start) len in
    if start < 0 || len < 1 || start + len > l then
      raise
      @@ Invalid_argument
           (Format.sprintf "start %d len %d poly size %d ctxt size %d" start len
              n ctxt_size);
    let res = Bls12_381.G1.(copy zero) in
    Stubs.pippenger res srs coefficients start len ctxt_size;
    res
end
